from collections import defaultdict
from datetime import datetime
from itertools import chain
from math import ceil
from typing import Optional, Iterable, Iterator, List, Dict

from maggma.builders import Builder
from maggma.core import Store
from maggma.utils import grouper

from emmet.core.qchem.task import TaskDocument
from emmet.core.qchem.molecule import MoleculeDoc, evaluate_lot
from emmet.core.molecules.vibration import VibrationDoc
from emmet.core.utils import jsanitize
from emmet.builders.settings import EmmetBuildSettings


__author__ = "Evan Spotte-Smith"

SETTINGS = EmmetBuildSettings()


class VibrationBuilder(Builder):
    """
    The VibrationBuilder extracts the highest-quality vibrational data from a
    MoleculeDoc (lowest electronic energy, highest level of theory for
    each solvent available).

    The process is as follows:
        1. Gather MoleculeDocs by formula
        2. For each doc, sort tasks by solvent
        3. For each solvent, grab the best TaskDoc (doc with vibrational
            information that has the highest level of theory with lowest
             electronic energy for the molecule)
        4. Convert TaskDoc to VibrationDoc

    Note that if no tasks associated with a given MoleculeDoc have vibrational
    data associated with them, then no VibrationDoc will be made for
    that molecule.
    """

    def __init__(
        self,
        tasks: Store,
        molecules: Store,
        vibes: Store,
        query: Optional[Dict] = None,
        settings: Optional[EmmetBuildSettings] = None,
        **kwargs,
    ):
        self.tasks = tasks
        self.molecules = molecules
        self.vibes = vibes
        self.query = query if query else dict()
        self.settings = EmmetBuildSettings.autoload(settings)
        self.kwargs = kwargs

        super().__init__(sources=[tasks, molecules], targets=[vibes], **kwargs)
        # Uncomment in case of issue with mrun not connecting automatically to collections
        # for i in [self.tasks, self.molecules, self.vibes]:
        #     try:
        #         i.connect()
        #     except Exception as e:
        #         print("Could not connect,", e)

    def ensure_indexes(self):
        """
        Ensures indices on the collections needed for building
        """

        # Basic search index for tasks
        self.tasks.ensure_index("task_id")
        self.tasks.ensure_index("last_updated")
        self.tasks.ensure_index("state")
        self.tasks.ensure_index("formula_alphabetical")

        # Search index for molecules
        self.molecules.ensure_index("molecule_id")
        self.molecules.ensure_index("last_updated")
        self.molecules.ensure_index("task_ids")
        self.molecules.ensure_index("formula_alphabetical")

        # Search index for vibrational properties
        self.vibes.ensure_index("molecule_id")
        self.vibes.ensure_index("task_id")
        self.vibes.ensure_index("solvent")
        self.vibes.ensure_index("lot_solvent")
        self.vibes.ensure_index("property_id")
        self.vibes.ensure_index("last_updated")
        self.vibes.ensure_index("formula_alphabetical")

    def prechunk(self, number_splits: int) -> Iterable[Dict]:  # pragma: no cover
        """Prechunk the builder for distributed computation"""

        temp_query = dict(self.query)
        temp_query["deprecated"] = False

        self.logger.info("Finding documents to process")
        all_mols = list(
            self.molecules.query(
                temp_query, [self.molecules.key, "formula_alphabetical"]
            )
        )

        processed_docs = set([e for e in self.vibes.distinct("molecule_id")])
        to_process_docs = {d[self.molecules.key] for d in all_mols} - processed_docs
        to_process_forms = {
            d["formula_alphabetical"]
            for d in all_mols
            if d[self.molecules.key] in to_process_docs
        }

        N = ceil(len(to_process_forms) / number_splits)

        for formula_chunk in grouper(to_process_forms, N):
            yield {"query": {"formula_alphabetical": {"$in": list(formula_chunk)}}}

    def get_items(self) -> Iterator[List[Dict]]:
        """
        Gets all items to process into vibration documents.
        This does no datetime checking; relying on on whether
        task_ids are included in the vibes Store

        Returns:
            generator or list relevant tasks and molecules to process into documents
        """

        self.logger.info("Vibration builder started")
        self.logger.info("Setting indexes")
        self.ensure_indexes()

        # Save timestamp to mark buildtime
        self.timestamp = datetime.utcnow()

        # Get all processed molecules
        temp_query = dict(self.query)
        temp_query["deprecated"] = False

        self.logger.info("Finding documents to process")
        all_mols = list(
            self.molecules.query(
                temp_query, [self.molecules.key, "formula_alphabetical"]
            )
        )

        processed_docs = set([e for e in self.vibes.distinct("molecule_id")])
        to_process_docs = {d[self.molecules.key] for d in all_mols} - processed_docs
        to_process_forms = {
            d["formula_alphabetical"]
            for d in all_mols
            if d[self.molecules.key] in to_process_docs
        }

        self.logger.info(f"Found {len(to_process_docs)} unprocessed documents")
        self.logger.info(f"Found {len(to_process_forms)} unprocessed formulas")

        # Set total for builder bars to have a total
        self.total = len(to_process_forms)

        for formula in to_process_forms:
            mol_query = dict(temp_query)
            mol_query["formula_alphabetical"] = formula
            molecules = list(self.molecules.query(criteria=mol_query))

            yield molecules

    def process_item(self, items: List[Dict]) -> List[Dict]:
        """
        Process the tasks into VibrationDocs

        Args:
            items List[Dict] : a list of MoleculeDocs in dict form

        Returns:
            [dict] : a list of new vibration docs
        """

        mols = [MoleculeDoc(**item) for item in items]
        formula = mols[0].formula_alphabetical
        mol_ids = [m.molecule_id for m in mols]
        self.logger.debug(f"Processing {formula} : {mol_ids}")

        vibe_docs = list()

        for mol in mols:
            vibe_entries = [
                e
                for e in mol.entries
                if e["charge"] == mol.charge
                and e["spin_multiplicity"] == mol.spin_multiplicity
                and e["output"].get("frequencies") is not None
            ]

            # Organize by solvent environment
            by_solvent = defaultdict(list)
            for entry in vibe_entries:
                by_solvent[entry["solvent"]].append(entry)

            for solvent, entries in by_solvent.items():
                # No documents with enthalpy and entropy
                if len(entries) == 0:
                    continue
                else:
                    best = sorted(
                        entries,
                        key=lambda x: (
                            sum(evaluate_lot(x["level_of_theory"])),
                            x["energy"],
                        ),
                    )[0]
                    task = best["task_id"]

                tdoc = self.tasks.query_one(
                    {
                        "task_id": task,
                        "formula_alphabetical": formula,
                        "orig": {"$exists": True},
                    }
                )

                if tdoc is None:
                    try:
                        tdoc = self.tasks.query_one(
                            {
                                "task_id": int(task),
                                "formula_alphabetical": formula,
                                "orig": {"$exists": True},
                            }
                        )
                    except ValueError:
                        tdoc = None

                if tdoc is None:
                    continue

                task_doc = TaskDocument(**tdoc)

                if task_doc is None:
                    continue

                vibe_doc = VibrationDoc.from_task(
                    task_doc, molecule_id=mol.molecule_id, deprecated=False
                )
                vibe_docs.append(vibe_doc)

        self.logger.debug(f"Produced {len(vibe_docs)} vibration docs for {formula}")

        return jsanitize([doc.model_dump() for doc in vibe_docs], allow_bson=True)

    def update_targets(self, items: List[List[Dict]]):
        """
        Inserts the new vibration docs into the vibes collection

        Args:
            items [[dict]]: A list of documents to update
        """

        docs = list(chain.from_iterable(items))  # type: ignore

        # Add timestamp
        for item in docs:
            item.update(
                {
                    "_bt": self.timestamp,
                }
            )

        molecule_ids = list({item["molecule_id"] for item in docs})

        if len(items) > 0:
            self.logger.info(f"Updating {len(docs)} vibration documents")
            self.vibes.remove_docs({self.vibes.key: {"$in": molecule_ids}})
            self.vibes.update(
                docs=docs,
                key=["molecule_id", "solvent"],
            )
        else:
            self.logger.info("No items to update")



import tempfile
import traceback
from math import ceil
from maggma.utils import grouper
from typing import Optional, Dict, List, Iterator

from abipy.dfpt.vsound import SoundVelocity as AbiSoundVelocity
from abipy.dfpt.ddb import DdbFile
from maggma.builders import Builder
from maggma.core import Store
from abipy.flowtk.tasks import TaskManager

from emmet.core.phonon import SoundVelocity
from emmet.core.utils import jsanitize


class SoundVelocityBuilder(Builder):
    def __init__(
        self,
        phonon_materials: Store,
        ddb_source: Store,
        sound_vel: Store,
        query: Optional[dict] = None,
        manager: Optional[TaskManager] = None,
        **kwargs
    ):
        """
        Creates a collection with the data of the sound velocities extracted from
        the phonon calculations.

        Args:
            phonon_materials (Store): source Store of phonon materials documents
                containing abinit_input and abinit_output.
            ddb_source (Store): source Store of ddb files. Matching the data in the materials Store.
            sound_vel (Store): target Store of the sound velocity
            query (dict): dictionary to limit materials to be analyzed
            manager (TaskManager): an instance of the abipy TaskManager. If None it will be
                generated from user configuration.
        """

        self.phonon_materials = phonon_materials
        self.ddb_source = ddb_source
        self.sound_vel = sound_vel
        self.query = query or {}

        if manager is None:
            self.manager = TaskManager.from_user_config()
        else:
            self.manager = manager

        super().__init__(
            sources=[phonon_materials, ddb_source], targets=[sound_vel], **kwargs
        )

    def prechunk(self, number_splits: int):  # pragma: no cover
        """
        Gets all materials that need sound velocity

        Returns:
            generator of materials to extract phonon sound velocity
        """

        # All relevant materials that have been updated since phonon props were last calculated
        q = dict(self.query)

        mats = self.sound_vel.newer_in(
            self.phonon_materials, exhaustive=True, criteria=q
        )

        N = ceil(len(mats) / number_splits)

        for mpid_chunk in grouper(mats, N):
            yield {"query": {self.phonon_materials.key: {"$in": list(mpid_chunk)}}}

    def get_items(self) -> Iterator[Dict]:
        """
        Gets all materials that need sound velocity.

        Returns:
            generator of materials to extract the sound velocity
        """

        self.logger.info("Sound Velocity Builder Started")

        self.logger.info("Setting indexes")
        self.ensure_indexes()

        # All relevant materials that have been updated since sound velocities were last calculated
        q = dict(self.query)
        mats = self.sound_vel.newer_in(
            self.phonon_materials, exhaustive=True, criteria=q
        )
        self.logger.info(
            "Found {} new materials for sound velocity data".format(len(mats))
        )

        # list of properties queried from the results DB
        # basic informations
        projection = {"mp_id": 1}
        # input data
        projection["abinit_input"] = 1
        # file ids to be fetched
        projection["abinit_output.ddb_id"] = 1

        for m in mats:
            item = self.phonon_materials.query_one(
                properties=projection, criteria={self.phonon_materials.key: m}
            )

            # Read the DDB file and pass as an object. Do not write here since in case of parallel
            # execution each worker will write its own file.
            ddb_data = self.ddb_source.query_one(
                criteria={"_id": item["abinit_output"]["ddb_id"]}
            )

            item["ddb_str"] = ddb_data["data"].decode("utf-8")

            yield item

    def process_item(self, item: Dict) -> Optional[Dict]:
        """
        Generates the sound velocity document from an item

        Args:
            item (dict): a dict extracted from the phonon calculations results.

        Returns:
            dict: a dict with phonon data
        """
        self.logger.debug("Processing sound velocity item for {}".format(item["mp_id"]))

        try:
            sound_vel_data = self.get_sound_vel(item)

            sv = SoundVelocity(
                material_id=item["mp_id"],
                structure=sound_vel_data["structure"],
                directions=sound_vel_data["directions"],
                labels=sound_vel_data["labels"],
                sound_velocities=sound_vel_data["sound_velocities"],
                mode_types=sound_vel_data["mode_types"],
            )

            self.logger.debug("Item generated for {}".format(item["mp_id"]))

            return jsanitize(sv.model_dump())
        except Exception:
            self.logger.warning(
                "Error generating the sound velocity for {}: {}".format(
                    item["mp_id"], traceback.format_exc()
                )
            )
            return None

    @staticmethod
    def get_sound_vel(item: Dict) -> Dict:
        """
        Runs anaddb and return the extracted data for the speed of sound.

        Args:
            item (dict): the item to process
        Returns:
            A dictionary with the sound velocity values
        """
        with tempfile.NamedTemporaryFile(
            mode="wt", suffix="_DDB", delete=True
        ) as ddb_file:
            ddb_file.write(item["ddb_str"])
            ngqpt = item["abinit_input"]["ngqpt"]
            sv = AbiSoundVelocity.from_ddb(
                ddb_file.name,
                ngqpt=ngqpt,
                num_points=20,
                qpt_norm=0.1,
                ignore_neg_freqs=True,
                directions=None,
            )

            ddb = DdbFile.from_string(item["ddb_str"])
            sv_data = dict(
                directions=sv.directions.tolist(),
                sound_velocities=sv.sound_velocities.tolist(),
                mode_types=sv.mode_types,
                labels=sv.labels,
                structure=ddb.structure,
            )

            return sv_data

    def update_targets(self, items: List[Dict]):
        """
        Inserts the new task_types into the task_types collection

        Args:
            items ([dict]): a list of dictionaries with sound velocities
                to update.
        """
        self.logger.debug("Start update_targets")
        items = list(filter(None, items))

        if len(items) > 0:
            self.logger.info("Updating {} sound velocity docs".format(len(items)))
            self.sound_vel.update(docs=items)
        else:
            self.logger.info("No items to update")

    def ensure_indexes(self):
        """
        Ensures indexes on the sound_vel collection.
        """

        # Search index for sound velocity
        self.sound_vel.ensure_index(self.sound_vel.key, unique=True)



from math import ceil
from typing import Dict, Optional, Iterator


import numpy as np
from maggma.builders import Builder
from maggma.core import Store
from maggma.utils import grouper
from pymatgen.core.structure import Structure

from emmet.core.polar import DielectricDoc
from emmet.core.utils import jsanitize


class DielectricBuilder(Builder):
    def __init__(
        self,
        materials: Store,
        tasks: Store,
        dielectric: Store,
        query: Optional[Dict] = None,
        **kwargs,
    ):
        self.materials = materials
        self.tasks = tasks
        self.dielectric = dielectric
        self.query = query or {}
        self.kwargs = kwargs

        self.materials.key = "material_id"
        self.tasks.key = "task_id"
        self.dielectric.key = "material_id"

        super().__init__(sources=[materials, tasks], targets=[dielectric], **kwargs)

    def prechunk(self, number_splits: int) -> Iterator[Dict]:  # pragma: no cover
        """
        Prechunk method to perform chunking by the key field
        """
        q = dict(self.query)

        keys = self.dielectric.newer_in(self.materials, criteria=q, exhaustive=True)

        N = ceil(len(keys) / number_splits)
        for split in grouper(keys, N):
            yield {"query": {self.materials.key: {"$in": list(split)}}}

    def get_items(self):
        """
        Gets all items to process

        Returns:
            generator or list relevant tasks and materials to process
        """

        self.logger.info("Dielectric Builder Started")

        q = dict(self.query)

        mat_ids = self.materials.distinct(self.materials.key, criteria=q)
        di_ids = self.dielectric.distinct(self.dielectric.key)

        mats_set = set(
            self.dielectric.newer_in(target=self.materials, criteria=q, exhaustive=True)
        ) | (set(mat_ids) - set(di_ids))

        mats = [mat for mat in mats_set]

        self.logger.info(
            "Processing {} materials for dielectric data".format(len(mats))
        )

        self.total = len(mats)

        for mat in mats:
            doc = self._get_processed_doc(mat)

            if doc is not None:
                yield doc
            else:
                pass

    def process_item(self, item):
        structure = Structure.from_dict(item["structure"])
        mpid = item["material_id"]
        origin_entry = {
            "name": "dielectric",
            "task_id": item["task_id"],
            "last_updated": item["task_updated"],
        }

        doc = DielectricDoc.from_ionic_and_electronic(
            structure=structure,
            material_id=mpid,
            origins=[origin_entry],
            deprecated=False,
            ionic=item["epsilon_ionic"],
            electronic=item["epsilon_static"],
            last_updated=item["updated_on"],
        )

        return jsanitize(doc.model_dump(), allow_bson=True)

    def update_targets(self, items):
        """
        Inserts the new dielectric docs into the dielectric collection
        """
        docs = list(filter(None, items))

        if len(docs) > 0:
            self.logger.info(f"Found {len(docs)} dielectric docs to update")
            self.dielectric.update(docs)
        else:
            self.logger.info("No items to update")

    def _get_processed_doc(self, mat):
        mat_doc = self.materials.query_one(
            {self.materials.key: mat},
            [
                self.materials.key,
                "structure",
                "task_types",
                "run_types",
                "deprecated_tasks",
                "last_updated",
            ],
        )

        task_types = mat_doc["task_types"].items()

        potential_task_ids = []

        for task_id, task_type in task_types:
            if task_type == "DFPT Dielectric":
                if task_id not in mat_doc["deprecated_tasks"]:
                    potential_task_ids.append(task_id)

        final_docs = []

        for task_id in potential_task_ids:
            task_query = self.tasks.query_one(
                properties=[
                    "last_updated",
                    "input.is_hubbard",
                    "orig_inputs.kpoints",
                    "orig_inputs.poscar.structure",
                    "input.parameters",
                    "input.structure",
                    "output.epsilon_static",
                    "output.epsilon_ionic",
                    "output.bandgap",
                ],
                criteria={self.tasks.key: str(task_id)},
            )

            if task_query["output"]["bandgap"] > 0:
                try:
                    structure = task_query["orig_inputs"]["poscar"]["structure"]
                except KeyError:
                    structure = task_query["input"]["structure"]

                is_hubbard = task_query["input"]["is_hubbard"]

                if (
                    task_query["orig_inputs"]["kpoints"]["generation_style"]
                    == "Monkhorst"
                    or task_query["orig_inputs"]["kpoints"]["generation_style"]
                    == "Gamma"
                ):
                    nkpoints = np.prod(
                        task_query["orig_inputs"]["kpoints"]["kpoints"][0], axis=0
                    )

                else:
                    nkpoints = task_query["orig_inputs"]["kpoints"]["nkpoints"]

                lu_dt = mat_doc["last_updated"]
                task_updated = task_query["last_updated"]

                final_docs.append(
                    {
                        "task_id": task_id,
                        "is_hubbard": int(is_hubbard),
                        "nkpoints": int(nkpoints),
                        "epsilon_static": task_query["output"]["epsilon_static"],
                        "epsilon_ionic": task_query["output"]["epsilon_ionic"],
                        "structure": structure,
                        "updated_on": lu_dt,
                        "task_updated": task_updated,
                        self.materials.key: mat_doc[self.materials.key],
                    }
                )

        if len(final_docs) > 0:
            sorted_final_docs = sorted(
                final_docs,
                key=lambda entry: (
                    entry["is_hubbard"],
                    entry["nkpoints"],
                    entry["updated_on"],
                ),
                reverse=True,
            )
            return sorted_final_docs[0]
        else:
            return None



from math import ceil
import warnings
from itertools import chain
from typing import Dict, Iterator, List, Optional, Set
from datetime import datetime

from maggma.core import Builder, Store
from maggma.stores import S3Store
from maggma.utils import grouper
from monty.json import MontyDecoder
from pymatgen.analysis.phase_diagram import PhaseDiagramError
from pymatgen.entries.computed_entries import ComputedStructureEntry

from emmet.builders.utils import HiddenPrints
from emmet.core.thermo import ThermoDoc, PhaseDiagramDoc
from emmet.core.utils import jsanitize


class ThermoBuilder(Builder):
    def __init__(
        self,
        thermo: Store,
        corrected_entries: Store,
        phase_diagram: Optional[Store] = None,
        query: Optional[Dict] = None,
        num_phase_diagram_eles: Optional[int] = None,
        chunk_size: int = 1000,
        **kwargs,
    ):
        """
        Calculates thermodynamic quantities for materials from phase
        diagram constructions

        Args:
            thermo (Store): Store of thermodynamic data such as formation
                energy and decomposition pathway
            corrected_entries (Store): Store of corrected entry data to use in thermo data and phase diagram
                construction. This is required and should be built with the CorrectedEntriesBuilder.
            phase_diagram (Store): Store of phase diagram data for each unique chemical system
            query (dict): dictionary to limit materials to be analyzed
            num_phase_diagram_eles (int): Maximum number of elements to use in phase diagram construction
                for data within the separate phase_diagram collection
            chunk_size (int): Size of chemsys chunks to process at any one time.
        """

        self.thermo = thermo
        self.query = query if query else {}
        self.corrected_entries = corrected_entries
        self.phase_diagram = phase_diagram
        self.num_phase_diagram_eles = num_phase_diagram_eles
        self.chunk_size = chunk_size
        self._completed_tasks: Set[str] = set()

        if self.thermo.key != "thermo_id":
            warnings.warn(
                f"Key for the thermo store is incorrect and has been changed from {self.thermo.key} to thermo_id!"
            )
            self.thermo.key = "thermo_id"

        if self.corrected_entries.key != "chemsys":
            warnings.warn(
                "Key for the corrected entries store is incorrect and has been changed "
                f"from {self.corrected_entries.key} to chemsys!"
            )
            self.corrected_entries.key = "chemsys"

        sources = [corrected_entries]
        targets = [thermo]

        if self.phase_diagram is not None:
            if self.phase_diagram.key != "phase_diagram_id":
                warnings.warn(
                    f"Key for the phase diagram store is incorrect and has been changed from {self.phase_diagram.key} to phase_diagram_id!"  # noqa: E501
                )
                self.phase_diagram.key = "phase_diagram_id"

            targets.append(phase_diagram)  # type: ignore

        super().__init__(
            sources=sources, targets=targets, chunk_size=chunk_size, **kwargs
        )

    def ensure_indexes(self):
        """
        Ensures indicies on the tasks and materials collections
        """

        # Search index for corrected_entries
        self.corrected_entries.ensure_index("chemsys")
        self.corrected_entries.ensure_index("last_updated")

        # Search index for thermo
        self.thermo.ensure_index("material_id")
        self.thermo.ensure_index("thermo_id")
        self.thermo.ensure_index("thermo_type")
        self.thermo.ensure_index("last_updated")

        # Search index for thermo
        self.thermo.ensure_index("material_id")
        self.thermo.ensure_index("thermo_id")
        self.thermo.ensure_index("thermo_type")
        self.thermo.ensure_index("last_updated")

        # Search index for phase_diagram
        if self.phase_diagram:
            coll = self.phase_diagram

            if isinstance(self.phase_diagram, S3Store):
                coll = self.phase_diagram.index

            coll.ensure_index("chemsys")
            coll.ensure_index("phase_diagram_id")

    def prechunk(self, number_splits: int) -> Iterator[Dict]:  # pragma: no cover
        to_process_chemsys = self._get_chemsys_to_process()

        N = ceil(len(to_process_chemsys) / number_splits)

        for chemsys_chunk in grouper(to_process_chemsys, N):
            yield {"query": {"chemsys": {"$in": list(chemsys_chunk)}}}

    def get_items(self) -> Iterator[List[Dict]]:
        """
        Gets whole chemical systems of entries to process
        """

        self.logger.info("Thermo Builder Started")

        self.logger.info("Setting indexes")
        self.ensure_indexes()

        to_process_chemsys = self._get_chemsys_to_process()

        self.logger.info(
            f"Found {len(to_process_chemsys)} chemical systems with new/updated materials to process"
        )
        self.total = len(to_process_chemsys)

        # Yield the chemical systems in order of increasing size
        # Will build them in a similar manner to fast Pourbaix
        for chemsys in sorted(
            to_process_chemsys, key=lambda x: len(x.split("-")), reverse=False
        ):
            corrected_entries = self.corrected_entries.query_one({"chemsys": chemsys})
            yield corrected_entries

    def process_item(self, item):
        if not item:
            return None

        pd_thermo_doc_pair_list = []

        for thermo_type, entry_list in item["entries"].items():
            if entry_list:
                entries = [
                    ComputedStructureEntry.from_dict(entry) for entry in entry_list
                ]
                chemsys = item["chemsys"]
                elements = chemsys.split("-")

                self.logger.debug(
                    f"Processing {len(entries)} entries for {chemsys} and thermo type {thermo_type}"
                )

                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    with HiddenPrints():
                        pd_thermo_doc_pair_list.append(
                            self._produce_pair(entries, thermo_type, elements)
                        )

        return pd_thermo_doc_pair_list

    def _produce_pair(self, pd_entries, thermo_type, elements):
        # Produce thermo and phase diagram pair

        try:
            # Obtain phase diagram
            pd = ThermoDoc.construct_phase_diagram(pd_entries)

            # Iterate through entry material IDs and construct list of thermo docs to update
            docs = ThermoDoc.from_entries(
                pd_entries, thermo_type, pd, use_max_chemsys=True, deprecated=False
            )

            pd_docs = [None]

            if self.phase_diagram:
                if (
                    self.num_phase_diagram_eles is None
                    or len(elements) <= self.num_phase_diagram_eles
                ):
                    chemsys = "-".join(sorted(set([e.symbol for e in pd.elements])))
                    pd_id = "{}_{}".format(chemsys, str(thermo_type))
                    pd_doc = PhaseDiagramDoc(
                        phase_diagram_id=pd_id,
                        chemsys=chemsys,
                        phase_diagram=pd,
                        thermo_type=thermo_type,
                    )

                    pd_data = jsanitize(pd_doc.model_dump(), allow_bson=True)

                    pd_docs = [pd_data]

            docs_pd_pair = (
                jsanitize([d.model_dump() for d in docs], allow_bson=True),
                pd_docs,
            )

            return docs_pd_pair

        except PhaseDiagramError as p:
            elsyms = []
            for e in pd_entries:
                elsyms.extend([el.symbol for el in e.composition.elements])

            self.logger.error(
                f"Phase diagram error in chemsys {'-'.join(sorted(set(elsyms)))}: {p}"
            )
            return (None, None)

    def update_targets(self, items):
        """
        Inserts the thermo and phase diagram docs into the thermo collection
        Args:
            items ([[tuple(List[dict],List[dict])]]): a list of a list of thermo and phase diagram dict pairs to update
        """

        thermo_docs = [pair[0] for pair_list in items for pair in pair_list]
        phase_diagram_docs = [pair[1] for pair_list in items for pair in pair_list]

        # flatten out lists
        thermo_docs = list(filter(None, chain.from_iterable(thermo_docs)))
        phase_diagram_docs = list(filter(None, chain.from_iterable(phase_diagram_docs)))

        # Check if already updated this run
        thermo_docs = [
            i for i in thermo_docs if i["thermo_id"] not in self._completed_tasks
        ]

        self._completed_tasks |= {i["thermo_id"] for i in thermo_docs}

        for item in thermo_docs:
            if isinstance(item["last_updated"], dict):
                item["last_updated"] = MontyDecoder().process_decoded(
                    item["last_updated"]
                )

        if self.phase_diagram:
            self.phase_diagram.update(phase_diagram_docs)

        if len(thermo_docs) > 0:
            self.logger.info(f"Updating {len(thermo_docs)} thermo documents")
            self.thermo.update(docs=thermo_docs, key=["thermo_id"])
        else:
            self.logger.info("No thermo items to update")

    def _get_chemsys_to_process(self):
        # Use last-updated to find new chemsys
        corrected_entries_chemsys_dates = {
            d[self.corrected_entries.key]: d[self.corrected_entries.last_updated_field]
            for d in self.corrected_entries.query(
                self.query,
                properties=[
                    self.corrected_entries.key,
                    self.corrected_entries.last_updated_field,
                ],
            )
        }

        thermo_chemsys_dates = {}
        for d in self.thermo.query(
            {"deprecated": False},
            properties=[self.corrected_entries.key, self.thermo.last_updated_field],
        ):
            entry = thermo_chemsys_dates.get(d[self.corrected_entries.key], None)
            if entry is None or d[self.thermo.last_updated_field] < entry:
                thermo_chemsys_dates[d[self.corrected_entries.key]] = d[
                    self.thermo.last_updated_field
                ]

        to_process_chemsys = [
            chemsys
            for chemsys in corrected_entries_chemsys_dates
            if (chemsys not in thermo_chemsys_dates)
            or (
                thermo_chemsys_dates[chemsys]
                < datetime.fromisoformat(corrected_entries_chemsys_dates[chemsys])
            )
        ]

        return to_process_chemsys



from emmet.builders.utils import (
    chemsys_permutations,
    maximal_spanning_non_intersecting_subsets,
    get_hop_cutoff,
    get_potcar_stats,
)
from pymatgen.analysis.diffusion.neb.full_path_mapper import MigrationGraph
from numpy.testing import assert_almost_equal
from monty.serialization import loadfn, dumpfn
from emmet.core.settings import EmmetSettings

import pytest


def test_maximal_spanning_non_intersecting_subsets():
    assert maximal_spanning_non_intersecting_subsets([{"A"}, {"A", "B"}]) == {
        frozenset(d) for d in [{"A"}, {"B"}]
    }

    assert maximal_spanning_non_intersecting_subsets([{"A", "B"}, {"A", "B", "C"}]) == {
        frozenset(d) for d in [{"A", "B"}, {"C"}]
    }

    assert maximal_spanning_non_intersecting_subsets(
        [{"A", "B"}, {"A", "B", "C"}, {"D"}]
    ) == {frozenset(d) for d in [{"A", "B"}, {"C"}, {"D"}]}


def test_chemsys_permutations(test_dir):
    assert len(chemsys_permutations("Sr")) == 1
    assert len(chemsys_permutations("Sr-Hf")) == 3
    assert len(chemsys_permutations("Sr-Hf-O")) == 7


def test_get_hop_cutoff(test_dir):
    spinel_mg = loadfn(test_dir / "mobility/migration_graph_spinel_MgMn2O4.json")
    nasicon_mg = loadfn(test_dir / "mobility/migration_graph_nasicon_MgV2(PO4)3.json")

    # tests for "min_distance" algorithm
    assert_almost_equal(
        get_hop_cutoff(spinel_mg.structure, "Mg", algorithm="min_distance"),
        1.95,
        decimal=2,
    )
    assert_almost_equal(
        get_hop_cutoff(nasicon_mg.structure, "Mg", algorithm="min_distance"),
        3.80,
        decimal=2,
    )

    # test for "hops_based" algorithm, terminated by number of unique hops condition
    d = get_hop_cutoff(spinel_mg.structure, "Mg", algorithm="hops_based")
    check_mg = MigrationGraph.with_distance(spinel_mg.structure, "Mg", d)
    assert_almost_equal(d, 4.18, decimal=2)
    assert len(check_mg.unique_hops) == 5

    # test for "hops_based" algorithm, terminated by the largest hop length condition
    d = get_hop_cutoff(nasicon_mg.structure, "Mg", algorithm="hops_based")
    check_mg = MigrationGraph.with_distance(nasicon_mg.structure, "Mg", d)
    assert_almost_equal(d, 4.59, decimal=2)
    assert len(check_mg.unique_hops) == 6


@pytest.mark.parametrize("method", ("potcar", "pymatgen", "stored"))
def test_get_potcar_stats(method: str, tmp_path):
    calc_type = EmmetSettings().VASP_DEFAULT_INPUT_SETS

    try:
        potcar_stats = get_potcar_stats(method=method)
    except Exception as exc:
        if "No POTCAR for" in str(exc):
            # No Potcar library available, skip test
            return
        else:
            raise exc

    # ensure that all calc types are included in potcar_stats
    assert potcar_stats.keys() == calc_type.keys()

    for calc_type in potcar_stats:
        # ensure that each entry has needed fields for both
        # legacy and modern potcar validation
        assert all(
            [
                set(entry) == set(["hash", "keywords", "titel", "stats"])
                for entry in entries
            ]
            for entries in potcar_stats[calc_type].values()
        )

    if method == "stored":
        new_stats_path = tmp_path / "_temp_potcar_stats.json"
        dumpfn(potcar_stats, new_stats_path)

        new_potcar_stats = get_potcar_stats(
            method="stored", path_to_stored_stats=new_stats_path
        )
        assert all(
            potcar_stats[calc_type] == new_potcar_stats[calc_type]
            for calc_type in potcar_stats
        )



from __future__ import annotations
from typing import Set, Union, Any, Literal, Optional
import sys
import os
from pathlib import Path
from gzip import GzipFile
import orjson
import json
from io import BytesIO
from monty.serialization import MontyDecoder
from botocore.exceptions import ClientError
from itertools import chain, combinations
from pymatgen.core import Structure
from pymatgen.analysis.diffusion.neb.full_path_mapper import MigrationGraph
from pymatgen.io.vasp.inputs import PotcarSingle

from emmet.builders.settings import EmmetBuildSettings


def maximal_spanning_non_intersecting_subsets(sets) -> Set[Set]:
    """
    Finds the maximal spanning non intersecting subsets of a group of sets
    This is usefull for parsing out the sandboxes and figuring out how to group
    and calculate these for thermo documents

    sets (set(frozenset)): sets of keys to subsect, expected as a set of frozensets
    """
    to_return_subsets = []

    # Find the overlapping portions and independent portions
    for subset in sets:
        for other_set in sets:
            subset = frozenset(subset.intersection(other_set)) or subset
        if subset:
            to_return_subsets.append(subset)

    # Remove accounted for elements and recurse on remaining sets
    accounted_elements = set(chain.from_iterable(to_return_subsets))
    sets = {frozenset(subset - accounted_elements) for subset in sets}
    sets = {subset for subset in sets if subset}

    if sets:
        to_return_subsets.extend(maximal_spanning_non_intersecting_subsets(sets))

    return set(to_return_subsets)


def chemsys_permutations(chemsys) -> Set:
    # Function to get all relevant chemical subsystems
    # e.g. for Li-Mn-O returns Li, Li-Mn, Li-Mn-O, Li-O, Mn, Mn-O, O
    elements = chemsys.split("-")
    return {
        "-".join(sorted(c))
        for c in chain(
            *[combinations(elements, i) for i in range(1, len(elements) + 1)]
        )
    }


def get_hop_cutoff(
    migration_graph_struct: Structure,
    mobile_specie: str,
    algorithm: str = "min_distance",
    min_hop_distance: float = 1,
    max_hop_distance: float = 7,
) -> Union[float, None]:
    """
    A function to get an appropriate hop distance cutoff for a given migration
    graph structure which can be used for MigrationGraph.with_distance()

    migration_graph_struct (Structure): the host structure with all working ion
        sites for consideration filled.
        Can get via MigrationGraph.get_structure_from_entries()
    mobile_specie (str): specifies the mobile specie in migration_graph_struct
        (e.g. "Li", "Mg").
    algorithm (str): specify the algorithm to use for getting the hop cutoff.
        "min_distance" =  incrementally increases the hop cutoff to get the
            minimum cutoff value that results in finding a path
        "hops_based" = incrementally increases the hop cutoff until either
            1) the largest hop length exceeds 2x the smallest hop length or
            2) the next incremental increase of the cutoff results in the total
            number of unique hops to exceeding more than 1.5x the previous
            number of unique hops
    """
    d = min_hop_distance
    mg = MigrationGraph.with_distance(
        structure=migration_graph_struct,
        migrating_specie=mobile_specie,
        max_distance=d,
    )
    mg.assign_cost_to_graph(["hop_distance"])
    paths = list(mg.get_path())

    if algorithm == "min_distance":
        while len(paths) == 0 and d < max_hop_distance:
            d = d * 1.1
            mg = MigrationGraph.with_distance(
                structure=migration_graph_struct,
                migrating_specie=mobile_specie,
                max_distance=d,
            )
            mg.assign_cost_to_graph(["hop_distance"])
            paths = list(mg.get_path())
        if d > max_hop_distance:
            return 0
        else:
            return d
    elif algorithm == "hops_based":
        # Thank you to Alex Nguyen Ngoc (alexnguyen0512) for their initial
        # work in developing this algorithm

        # incrementally increase d until at least one path is found
        # this sets d at the minimum distance that results in a path
        while len(paths) == 0 and d < max_hop_distance:
            d = d * 1.1
            mg = MigrationGraph.with_distance(
                structure=migration_graph_struct,
                migrating_specie=mobile_specie,
                max_distance=d,
            )
            mg.assign_cost_to_graph(["hop_distance"])
            paths = list(mg.get_path())
        if d > max_hop_distance:
            return 0

        # once a path is found, get the smallest hop distance and
        # the number of unique hops
        smallest_hop_distance = min(
            [v["hop_distance"] for v in mg.unique_hops.values()]
        )
        num_unique_hops = len(mg.unique_hops)

        # continually incrementally increase d until either...
        # 1) the largest hop distance exceeds 2x the smallest hop distance
        # 2) the number of unique hops spikes (increases by more than 1.5x)
        while (
            max([v["hop_distance"] for v in mg.unique_hops.values()])
            <= smallest_hop_distance * 2
            and d < max_hop_distance
        ):
            d = d * 1.1
            mg = MigrationGraph.with_distance(
                structure=migration_graph_struct,
                migrating_specie=mobile_specie,
                max_distance=d,
            )

            # check for spike in number of hops
            if len(mg.unique_hops) >= 1.5 * num_unique_hops and num_unique_hops >= 4:
                return d / 1.1  # return value of d before the number of hops spiked

            num_unique_hops = len(mg.unique_hops)

        return d

    else:
        return None


def query_open_data(
    bucket: str,
    prefix: str,
    key: str,
    monty_decode: bool = True,
    s3_resource: Any = None,
) -> Union[dict, None]:
    """Query a Materials Project AWS S3 Open Data bucket directly with boto3

    Args:
        bucket (str): Materials project bucket name
        prefix (str): Full set of file prefixes
        key (str): Key for file
        monty_decode (bool): Whether to monty decode or keep as dictionary. Defaults to True.
        s3_resource (Optional[Any]): S3 resource. One will be instantiated if none are provided

    Returns:
        dict: MontyDecoded data or None
    """

    def decode(content, monty_decode):
        if monty_decode:
            result = MontyDecoder().decode(content)
        else:
            result = orjson.loads(content)
        return result

    try:
        file_key = f"{prefix}/{key}.json.gz"
        ref = s3_resource.Object(bucket, file_key)  # type: ignore
        bytes = ref.get()["Body"]  # type: ignore

        with GzipFile(fileobj=bytes, mode="r") as gzipfile:
            content = gzipfile.read()

        try:
            result = decode(content, monty_decode)
        except (orjson.JSONDecodeError, json.JSONDecodeError):
            try:
                with GzipFile(fileobj=BytesIO(content), mode="r") as gzipfile_nested:
                    result = decode(gzipfile_nested.read(), monty_decode)
            except Exception:
                print(f"Issue decoding {file_key} from bucket {bucket}")
                return None
    except ClientError:
        return None

    return result


# From: https://stackoverflow.com/a/45669280
class HiddenPrints:
    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, "w")

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout


def get_potcar_stats(
    method: Literal["potcar", "pymatgen", "stored"] = "potcar",
    path_to_stored_stats: Optional[Union[str, os.PathLike, Path]] = None,
) -> dict[str, Any]:
    """
    Get the POTCAR stats used in MP calculations to validate POTCARs.

    Args:
        method : Literal[str : "potcar","pymatgen","stored"] = "potcar"
            Method to generate the POTCAR stats:
            - "potcar": regenerate stats from a user's POTCAR library.
            - "pymatgen": regenerate stats from the stored pymatgen
              summary stats dict. This has the downside of the possibility
              of finding multiple matching POTCAR stats for older POTCAR
              releases. As of 25 March, 2024, it does not appear that the
              MP POTCARs have duplicates
            - "stored": load a stored dict of POTCAR stats.
        path_to_stored_stats : str, os.Pathlike, Path, or None
            If a str, the path to the stored summary stats file.
            If None, defaults to
              `importlib.resources.file("emmet.builders.vasp") / "mp_potcar_stats.json.gz"`
    Returns:
        dict, of POTCAR summary stats.
    """
    default_settings = EmmetBuildSettings()

    stats: dict[str, dict] = {}  # type: ignore

    if method == "stored":
        from monty.serialization import loadfn

        if path_to_stored_stats is None:
            from importlib.resources import files

            path_to_stored_stats = str(
                files("emmet.builders.vasp") / "mp_potcar_stats.json.gz"
            )
        return loadfn(path_to_stored_stats)

    for (
        calc_type,
        input_set,
    ) in default_settings.VASP_DEFAULT_INPUT_SETS.items():
        _input = input_set()

        stats[calc_type] = {}
        functional = _input._config_dict["POTCAR_FUNCTIONAL"]

        for potcar_symbol in _input.CONFIG["POTCAR"].values():
            if method == "potcar":
                potcar = PotcarSingle.from_symbol_and_functional(
                    symbol=potcar_symbol, functional=functional
                )
                summary_stats = potcar._summary_stats.copy()
                # fallback method for validation - use header hash and symbol
                # note that the potcar_spec assigns PotcarSingle.symbol to "titel"
                # whereas the ***correct*** field is `header`
                summary_stats["titel"] = potcar.header
                summary_stats["hash"] = potcar.md5_header_hash
                summary_stats = [summary_stats]

            elif method == "pymatgen":
                summary_stats = []
                for _, entries in PotcarSingle._potcar_summary_stats[
                    functional
                ].items():
                    summary_stats += [
                        {**entry, "titel": None, "hash": None}
                        for entry in entries
                        if entry["symbol"] == potcar_symbol
                    ]

            stats[calc_type].update({potcar_symbol: summary_stats})

    return stats



import pytest
from maggma.stores import JSONStore, MemoryStore

from emmet.builders.settings import EmmetBuildSettings
from emmet.builders.vasp.task_validator import TaskValidator

intermediate_stores = ["validation"]


@pytest.fixture(scope="session")
def tasks_store(test_dir):
    return JSONStore(test_dir / "test_si_tasks.json.gz")


@pytest.fixture(scope="session")
def validation_store():
    return MemoryStore()


def test_validator(tasks_store, validation_store):
    settings = EmmetBuildSettings(VASP_VALIDATE_POTCAR_STATS=False)
    builder = TaskValidator(
        tasks=tasks_store, task_validation=validation_store, settings=settings
    )
    builder.run()
    assert validation_store.count() == tasks_store.count()
    assert validation_store.count({"valid": True}) == tasks_store.count()
    assert all(list(d["run_type"] == "GGA" for d in validation_store.query()))



from collections import defaultdict
from math import ceil
import itertools
import re
import boto3
import numpy as np
from botocore.handlers import disable_signing
from maggma.builders import Builder
from maggma.utils import grouper
from pymatgen.analysis.magnetism.analyzer import CollinearMagneticStructureAnalyzer
from pymatgen.core import Structure
from pymatgen.electronic_structure.core import Spin
from pymatgen.electronic_structure.bandstructure import BandStructureSymmLine
from pymatgen.electronic_structure.dos import CompleteDos
from pymatgen.symmetry.bandstructure import HighSymmKpath
from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.io.vasp.sets import MPStaticSet

from emmet.core.settings import EmmetSettings
from emmet.core.electronic_structure import ElectronicStructureDoc
from emmet.core.utils import jsanitize

from emmet.builders.utils import query_open_data

SETTINGS = EmmetSettings()


class ElectronicStructureBuilder(Builder):
    def __init__(
        self,
        tasks,
        materials,
        electronic_structure,
        bandstructure_fs,
        dos_fs,
        chunk_size=10,
        query=None,
        **kwargs,
    ):
        """
        Creates an electronic structure collection from a tasks collection,
        the associated band structures and density of states file store collections,
        and the materials collection.

        Individual bandstructures for each of the three conventions are generated.

        tasks (Store): Store of task documents
        materials (Store): Store of materials documents
        electronic_structure (Store): Store of electronic structure summary data documents
        bandstructure_fs (Store, str): Store of bandstructures, or S3 URL string with prefix
            (e.g. s3://materialsproject-parsed/bandstructures).
        dos_fs (Store, str): Store of DOS, or S3 URL string with bucket and prefix
            (e.g. s3://materialsproject-parsed/dos).
        chunk_size (int): Chunk size to use for processing. Defaults to 10.
        query (dict): Dictionary to limit materials to be analyzed
        """

        self.tasks = tasks
        self.materials = materials
        self.electronic_structure = electronic_structure
        self.bandstructure_fs = bandstructure_fs
        self.dos_fs = dos_fs
        self.chunk_size = chunk_size
        self.query = query if query else {}

        self._s3_resource = None

        sources = [tasks, materials]

        fs_stores = [bandstructure_fs, dos_fs]

        for store in fs_stores:
            if isinstance(store, str):
                if not re.match("^s3://.*", store):
                    raise ValueError(
                        "Please provide an S3 URL "
                        "in the format s3://{bucket_name}/{prefix}"
                    )

                if self._s3_resource is None:
                    self._s3_resource = boto3.resource("s3")
                    self._s3_resource.meta.client.meta.events.register(
                        "choose-signer.s3.*", disable_signing
                    )

            else:
                sources.append(store)

        super().__init__(
            sources=sources,
            targets=[electronic_structure],
            chunk_size=chunk_size,
            **kwargs,
        )

    def prechunk(self, number_splits: int):  # pragma: no cover
        """
        Prechunk method to perform chunking by the key field
        """
        q = dict(self.query)

        keys = self.electronic_structure.newer_in(
            self.materials, criteria=q, exhaustive=True
        )

        N = ceil(len(keys) / number_splits)
        for split in grouper(keys, N):
            yield {"query": {self.materials.key: {"$in": list(split)}}}

    def get_items(self):
        """
        Gets all items to process

        Returns:
            generator or list relevant tasks and materials to process
        """

        self.logger.info("Electronic Structure Builder Started")

        q = dict(self.query)

        mat_ids = self.materials.distinct(self.materials.key, criteria=q)
        es_ids = self.electronic_structure.distinct(self.electronic_structure.key)

        mats_set = set(
            self.electronic_structure.newer_in(
                target=self.materials, criteria=q, exhaustive=True
            )
        ) | (set(mat_ids) - set(es_ids))

        mats = [mat for mat in mats_set]

        self.logger.info(
            "Processing {} materials for electronic structure".format(len(mats))
        )

        self.total = len(mats)

        for mat in mats:
            mat = self._update_materials_doc(mat)
            yield mat

    def process_item(self, mat):
        """
        Process the band structures and dos data.

        Args:
            mat (dict): material document

        Returns:
            (dict): electronic_structure document
        """

        structure = Structure.from_dict(mat["structure"])

        self.logger.info("Processing: {}".format(mat[self.materials.key]))

        dos = None
        bs = {}
        structures = {}

        for bs_type, bs_entry in mat["bandstructure"].items():
            if bs_entry.get("object", None) is not None:
                bs[bs_type] = (
                    {
                        bs_entry["task_id"]: BandStructureSymmLine.from_dict(
                            bs_entry["object"]
                        )
                    }
                    if bs_entry
                    else None
                )

                structures[bs_entry["task_id"]] = bs_entry["output_structure"]

        if mat["dos"]:
            if mat["dos"]["object"] is not None:
                self.logger.info("Processing density of states")
                dos = {
                    mat["dos"]["task_id"]: CompleteDos.from_dict(mat["dos"]["object"])
                }

                structures[mat["dos"]["task_id"]] = mat["dos"]["output_structure"]

        if bs:
            self.logger.info(
                "Processing band structure types: {}".format(
                    [bs_type for bs_type, bs_entry in bs.items() if bs_entry]
                )
            )

        # Default summary data
        d = dict(
            material_id=mat[self.materials.key],
            deprecated=mat["deprecated"],
            task_id=mat["other"]["task_id"],
            meta_structure=structure,
            band_gap=mat["other"]["band_gap"],
            cbm=mat["other"]["cbm"],
            vbm=mat["other"]["vbm"],
            efermi=mat["other"]["efermi"],
            is_gap_direct=mat["other"]["is_gap_direct"],
            is_metal=mat["other"]["is_metal"],
            magnetic_ordering=mat["other"]["magnetic_ordering"],
            origins=mat["origins"],
            warnings=[],
        )

        # Eigenvalue band property checks
        eig_values = mat["other"].get("eigenvalue_band_properties", None)

        if eig_values is not None:
            if not np.isclose(
                mat["other"]["band_gap"], eig_values["bandgap"], atol=0.2, rtol=0.0
            ):
                d["warnings"].append(
                    "Regular parsed band gap and band gap from eigenvalue_band_properties do not agree. "
                    "Using data from eigenvalue_band_properties where appropriate."
                )

                d["band_gap"] = eig_values["bandgap"]
                d["cbm"] = eig_values["cbm"]
                d["vbm"] = eig_values["vbm"]
                d["is_gap_direct"] = eig_values["is_gap_direct"]
                d["is_metal"] = (
                    True if np.isclose(d["band_gap"], 0.0, atol=0.01, rtol=0) else False
                )

        if dos is None:
            doc = ElectronicStructureDoc.from_structure(**d)

        else:
            try:
                doc = ElectronicStructureDoc.from_bsdos(
                    material_id=mat[self.materials.key],
                    structures=structures,
                    dos=dos,
                    is_gap_direct=d["is_gap_direct"],
                    is_metal=d["is_metal"],
                    deprecated=d["deprecated"],
                    origins=d["origins"],
                    **bs,
                )
                doc = self._bsdos_checks(doc, dos[mat["dos"]["task_id"]], structures)

            except Exception:
                d["warnings"].append(
                    "Band structure and/or data exists but an error occured while processing."
                )
                doc = ElectronicStructureDoc.from_structure(**d)

        # Magnetic ordering check
        mag_orderings = {}
        if doc.bandstructure is not None:
            mag_orderings.update(
                {
                    bs_summary.task_id: bs_summary.magnetic_ordering
                    for bs_type, bs_summary in doc.bandstructure
                    if bs_summary is not None
                }
            )

        if doc.dos is not None:
            dos_dict = doc.dos.model_dump()
            mag_orderings.update(
                {dos_dict["total"][Spin.up]["task_id"]: dos_dict["magnetic_ordering"]}
            )

        for task_id, ordering in mag_orderings.items():
            if doc.magnetic_ordering != ordering:
                doc.warnings.append(
                    f"Summary data magnetic ordering does not agree with the ordering from {task_id}"
                )

        # LMAXMIX check, VASP default is 2
        expected_lmaxmix = MPStaticSet(structure).incar.get("LMAXMIX", 2)
        if mat["dos"] and mat["dos"]["lmaxmix"] != expected_lmaxmix:
            doc.warnings.append(
                "An incorrect calculation parameter may lead to errors in the band gap of "
                f"0.1-0.2 eV (LMAXIX is {mat['dos']['lmaxmix']} and should be {expected_lmaxmix} for "
                f"{mat['dos']['task_id']}). A correction calculation is planned."
            )

        for bs_type, bs_entry in mat["bandstructure"].items():
            if bs_entry["lmaxmix"] != expected_lmaxmix:
                doc.warnings.append(
                    "An incorrect calculation parameter may lead to errors in the band gap of "
                    f"0.1-0.2 eV (LMAXIX is {bs_entry['lmaxmix']} and should be {expected_lmaxmix} for "
                    f"{bs_entry['task_id']}). A correction calculation is planned."
                )

        return doc.model_dump()

    def update_targets(self, items):
        """
        Inserts electronic structure documents into the electronic_structure collection

        Args:
            items ([Dict]): A list of ElectronicStructureDoc dictionaries to update
        """

        items = list(filter(None, items))

        if len(items) > 0:
            self.logger.info("Updating {} electronic structure docs".format(len(items)))
            self.electronic_structure.update(docs=jsanitize(items, allow_bson=True))
        else:
            self.logger.info("No electronic structure docs to update")

    def _bsdos_checks(self, doc, dos, structures):
        # Band gap difference check for uniform and line-mode calculations
        bgap_diff = []
        for bs_type, bs_summary in doc.bandstructure:
            if bs_summary is not None:
                bgap_diff.append(doc.band_gap - bs_summary.band_gap)

        if dos is not None:
            bgap_diff.append(doc.band_gap - dos.get_gap())

        if any(abs(gap) > 0.25 for gap in bgap_diff):
            if doc.warnings is None:
                doc.warnings = []
            doc.warnings.append(
                "Absolute difference between blessed band gap and at least one "
                "line-mode or uniform calculation band gap is larger than 0.25 eV."
            )

        # Line-mode and uniform structure primitive checks

        pair_list = []
        for task_id, struct in structures.items():
            pair_list.append((task_id, struct))

            struct_prim = SpacegroupAnalyzer(struct).get_primitive_standard_structure(
                international_monoclinic=False
            )

            if not np.allclose(
                struct.lattice.matrix, struct_prim.lattice.matrix, atol=1e-3
            ):
                if doc.warnings is None:
                    doc.warnings = []

                if np.isclose(struct_prim.volume, struct.volume, atol=5, rtol=0):
                    doc.warnings.append(
                        f"The input structure for {task_id} is primitive but may not exactly match the "
                        f"standard primitive setting."
                    )
                else:
                    doc.warnings.append(
                        f"The input structure for {task_id} does not match the expected standard primitive"
                    )

        # Check line-mode and uniform for same structure
        sm = StructureMatcher()
        for pair in itertools.combinations(pair_list, 2):
            if not sm.fit(pair[0][1], pair[1][1]):
                if doc.warnings is None:
                    doc.warnings = []

                doc.warnings.append(
                    f"The input structures between bandstructure calculations {pair[0][0]} and {pair[1][0]} "
                    f"are not equivalent"
                )

        return doc

    def _update_materials_doc(self, mat_id):
        # find bs type for each task in task_type and store each different bs object

        mat = self.materials.query_one(
            properties=[
                self.materials.key,
                "structure",
                "inputs",
                "task_types",
                "deprecated",
                self.materials.last_updated_field,
            ],
            criteria={self.materials.key: mat_id},
        )

        mat["dos"] = {}
        mat["bandstructure"] = defaultdict(dict)
        mat["other"] = {}

        bs_calcs = defaultdict(list)
        dos_calcs = []
        other_calcs = []

        for task_id in mat["task_types"].keys():
            # Handle all line-mode tasks
            if "NSCF Line" in mat["task_types"][task_id]:
                bs_type = None

                task_query = self.tasks.query_one(
                    properties=[
                        "calcs_reversed",
                        "last_updated",
                        "input.is_hubbard",
                        "input.incar",
                        "orig_inputs.kpoints",
                        "input.parameters",
                        "output.structure",
                    ],
                    criteria={"task_id": str(task_id)},
                )

                fs_id = str(
                    task_query["calcs_reversed"][0].get("bandstructure_fs_id", None)
                )

                if fs_id is not None:
                    structure = Structure.from_dict(task_query["output"]["structure"])

                    kpoints = task_query["orig_inputs"]["kpoints"]

                    labels_dict = {
                        label: point
                        for label, point in zip(kpoints["labels"], kpoints["kpoints"])
                        if label is not None
                    }

                    try:
                        bs_type = self._obtain_path_type(labels_dict, structure)
                    except Exception:
                        bs_type = None

                    if bs_type is None:
                        if isinstance(self.bandstructure_fs, str):
                            _, _, bucket, prefix = self.bandstructure_fs.strip(
                                "/"
                            ).split("/")

                            bs_dict = query_open_data(
                                bucket,
                                prefix,
                                task_id,
                                monty_decode=False,
                                s3_resource=self._s3_resource,
                            )
                        else:
                            bs_dict = self.bandstructure_fs.query_one(
                                {self.bandstructure_fs.key: str(task_id)}
                            )

                        if bs_dict is not None:
                            bs = BandStructureSymmLine.from_dict(bs_dict["data"])

                            labels_dict = {
                                label: kpoint.frac_coords
                                for label, kpoint in bs.labels_dict.items()
                            }

                            try:
                                bs_type = self._obtain_path_type(
                                    labels_dict, bs.structure
                                )
                            except Exception:
                                bs_type = None

                        # Clear bs data
                        bs = None
                        bs_dict = None

                    is_hubbard = task_query["input"]["is_hubbard"]
                    lmaxmix = task_query["input"]["incar"].get(
                        "LMAXMIX", 2
                    )  # VASP default is 2, alternatively could project `parameters`
                    nkpoints = task_query["orig_inputs"]["kpoints"]["nkpoints"]
                    lu_dt = task_query["last_updated"]

                    if bs_type is not None:
                        bs_calcs[bs_type].append(
                            {
                                "fs_id": fs_id,
                                "task_id": task_id,
                                "is_hubbard": int(is_hubbard),
                                "lmaxmix": lmaxmix,
                                "nkpoints": int(nkpoints),
                                "updated_on": lu_dt,
                                "output_structure": structure,
                                "labels_dict": labels_dict,
                            }
                        )

            # Handle uniform tasks
            if "NSCF Uniform" in mat["task_types"][task_id]:
                task_query = self.tasks.query_one(
                    properties=[
                        "calcs_reversed",
                        "last_updated",
                        "input.is_hubbard",
                        "input.incar",
                        "orig_inputs.kpoints",
                        "input.parameters",
                        "output.structure",
                    ],
                    criteria={"task_id": str(task_id)},
                )

                fs_id = str(task_query["calcs_reversed"][0].get("dos_fs_id", None))

                if fs_id is not None:
                    lmaxmix = task_query["input"]["incar"].get(
                        "LMAXMIX", 2
                    )  # VASP default is 2, alternatively could project `parameters`

                    is_hubbard = task_query["input"]["is_hubbard"]

                    structure = Structure.from_dict(task_query["output"]["structure"])

                    if (
                        task_query["orig_inputs"]["kpoints"]["generation_style"]
                        == "Monkhorst"
                        or task_query["orig_inputs"]["kpoints"]["generation_style"]
                        == "Gamma"
                    ):
                        nkpoints = np.prod(
                            task_query["orig_inputs"]["kpoints"]["kpoints"][0], axis=0
                        )

                    else:
                        nkpoints = task_query["orig_inputs"]["kpoints"]["nkpoints"]

                    nedos = task_query["input"]["parameters"]["NEDOS"]
                    lu_dt = task_query["last_updated"]

                    dos_calcs.append(
                        {
                            "fs_id": fs_id,
                            "task_id": task_id,
                            "is_hubbard": int(is_hubbard),
                            "lmaxmix": lmaxmix,
                            "nkpoints": int(nkpoints),
                            "nedos": int(nedos),
                            "updated_on": lu_dt,
                            "output_structure": structure,
                        }
                    )

            # Handle static and structure opt tasks
            if "Static" or "Structure Optimization" in mat["task_types"][task_id]:
                task_query = self.tasks.query_one(
                    properties=[
                        "last_updated",
                        "input.is_hubbard",
                        "orig_inputs.kpoints",
                        "calcs_reversed",
                        "output.structure",
                    ],
                    criteria={"task_id": str(task_id)},
                )

                structure = Structure.from_dict(task_query["output"]["structure"])

                other_mag_ordering = CollinearMagneticStructureAnalyzer(
                    structure
                ).ordering

                is_hubbard = task_query["input"]["is_hubbard"]

                last_calc = task_query["calcs_reversed"][-1]

                if (
                    last_calc["input"]["kpoints"]["generation_style"] == "Monkhorst"
                    or last_calc["input"]["kpoints"]["generation_style"] == "Gamma"
                ):
                    nkpoints = np.prod(
                        last_calc["input"]["kpoints"]["kpoints"][0], axis=0
                    )
                else:
                    nkpoints = last_calc["input"]["kpoints"]["nkpoints"]

                lu_dt = task_query["last_updated"]

                other_calcs.append(
                    {
                        "is_static": True
                        if "Static" in mat["task_types"][task_id]
                        else False,
                        "task_id": task_id,
                        "is_hubbard": int(is_hubbard),
                        "nkpoints": int(nkpoints),
                        "magnetic_ordering": other_mag_ordering,
                        "updated_on": lu_dt,
                        "calcs_reversed": task_query["calcs_reversed"],
                    }
                )

        updated_materials_doc = self._obtain_blessed_calculations(
            mat, bs_calcs, dos_calcs, other_calcs
        )

        return updated_materials_doc

    def _obtain_blessed_calculations(
        self, materials_doc, bs_calcs, dos_calcs, other_calcs
    ):
        bs_types = ["setyawan_curtarolo", "hinuma", "latimer_munro"]

        materials_doc["origins"] = []

        for bs_type in bs_types:
            # select "blessed" bs of each type
            if bs_calcs[bs_type]:
                sorted_bs_data = sorted(
                    bs_calcs[bs_type],
                    key=lambda entry: (
                        entry["is_hubbard"],
                        entry["nkpoints"],
                        entry["updated_on"],
                    ),
                    reverse=True,
                )

                materials_doc["bandstructure"][bs_type]["task_id"] = sorted_bs_data[0][
                    "task_id"
                ]

                materials_doc["bandstructure"][bs_type]["lmaxmix"] = sorted_bs_data[0][
                    "lmaxmix"
                ]
                if isinstance(self.bandstructure_fs, str):
                    _, _, bucket, prefix = self.bandstructure_fs.strip("/").split("/")
                    bs_obj = query_open_data(
                        bucket,
                        prefix,
                        sorted_bs_data[0]["task_id"],
                        monty_decode=False,
                        s3_resource=self._s3_resource,
                    )
                else:
                    bs_obj = self.bandstructure_fs.query_one(
                        criteria={"fs_id": sorted_bs_data[0]["fs_id"]}
                    )

                materials_doc["bandstructure"][bs_type]["object"] = (
                    bs_obj["data"] if bs_obj is not None else None
                )

                materials_doc["bandstructure"][bs_type][
                    "output_structure"
                ] = sorted_bs_data[0]["output_structure"]

                materials_doc["origins"].append(
                    {
                        "name": bs_type,
                        "task_id": sorted_bs_data[0]["task_id"],
                        "last_updated": sorted_bs_data[0]["updated_on"],
                    }
                )

        if dos_calcs:
            sorted_dos_data = sorted(
                dos_calcs,
                key=lambda entry: (
                    entry["is_hubbard"],
                    entry["nkpoints"],
                    entry["nedos"],
                    entry["updated_on"],
                ),
                reverse=True,
            )

            materials_doc["dos"]["task_id"] = sorted_dos_data[0]["task_id"]

            materials_doc["dos"]["lmaxmix"] = sorted_dos_data[0]["lmaxmix"]

            if isinstance(self.bandstructure_fs, str):
                _, _, bucket, prefix = self.dos_fs.strip("/").split("/")
                dos_obj = query_open_data(
                    bucket,
                    prefix,
                    sorted_dos_data[0]["task_id"],
                    monty_decode=False,
                    s3_resource=self._s3_resource,
                )
            else:
                dos_obj = self.dos_fs.query_one(
                    criteria={"fs_id": sorted_dos_data[0]["fs_id"]}
                )

            materials_doc["dos"]["object"] = (
                dos_obj["data"] if dos_obj is not None else None
            )

            materials_doc["dos"]["output_structure"] = sorted_dos_data[0][
                "output_structure"
            ]

            materials_doc["origins"].append(
                {
                    "name": "dos",
                    "task_id": sorted_dos_data[0]["task_id"],
                    "last_updated": sorted_dos_data[0]["updated_on"],
                }
            )

        if other_calcs:
            sorted_other_data = sorted(
                other_calcs,
                key=lambda entry: (
                    entry["is_static"],
                    entry["is_hubbard"],
                    entry["nkpoints"],
                    entry["updated_on"],
                ),
                reverse=True,
            )

            materials_doc["other"]["task_id"] = str(sorted_other_data[0]["task_id"])

            task_output_data = sorted_other_data[0]["calcs_reversed"][-1]["output"]
            materials_doc["other"]["band_gap"] = task_output_data["bandgap"]
            materials_doc["other"]["magnetic_ordering"] = sorted_other_data[0][
                "magnetic_ordering"
            ]
            materials_doc["other"]["last_updated"] = sorted_other_data[0]["updated_on"]

            materials_doc["other"]["is_metal"] = (
                materials_doc["other"]["band_gap"] == 0.0
            )

            materials_doc["origins"].append(
                {
                    "name": "electronic_structure",
                    "task_id": sorted_other_data[0]["task_id"],
                    "last_updated": sorted_other_data[0]["updated_on"],
                }
            )

            for prop in [
                "efermi",
                "cbm",
                "vbm",
                "is_gap_direct",
                "is_metal",
                "eigenvalue_band_properties",
            ]:
                # First try other calcs_reversed entries if properties are not found in last
                if prop not in task_output_data:
                    for calc in sorted_other_data[0]["calcs_reversed"]:
                        if calc["output"].get(prop, None) is not None:
                            materials_doc["other"][prop] = calc["output"][prop]
                else:
                    materials_doc["other"][prop] = task_output_data[prop]

        return materials_doc

    @staticmethod
    def _obtain_path_type(
        labels_dict,
        structure,
        symprec=SETTINGS.SYMPREC,
        angle_tolerance=SETTINGS.ANGLE_TOL,
        atol=1e-5,
    ):
        bs_type = None

        if any([label.islower() for label in labels_dict]):
            bs_type = "latimer_munro"
        else:
            for ptype in ["setyawan_curtarolo", "hinuma"]:
                hskp = HighSymmKpath(
                    structure,
                    has_magmoms=False,
                    magmom_axis=None,
                    path_type=ptype,
                    symprec=symprec,
                    angle_tolerance=angle_tolerance,
                    atol=atol,
                )
                hs_labels_full = hskp.kpath["kpoints"]
                hs_path_uniq = set(
                    [label for segment in hskp.kpath["path"] for label in segment]
                )

                hs_labels = {
                    k: hs_labels_full[k] for k in hs_path_uniq if k in hs_path_uniq
                }

                shared_items = {
                    k: labels_dict[k]
                    for k in labels_dict
                    if k in hs_labels
                    and np.allclose(labels_dict[k], hs_labels[k], atol=1e-3)
                }

                if len(shared_items) == len(labels_dict) and len(shared_items) == len(
                    hs_labels
                ):
                    bs_type = ptype

        return bs_type



import tempfile
import os
from math import ceil
from emmet.builders.settings import EmmetBuildSettings
import numpy as np
from typing import Optional, Dict, List, Iterator, Tuple
from maggma.utils import grouper

from pymatgen.phonon.bandstructure import PhononBandStructureSymmLine
from pymatgen.phonon.dos import CompletePhononDos
from pymatgen.phonon.ir_spectra import IRDielectricTensor
from pymatgen.core.structure import Structure
from pymatgen.io.abinit.abiobjects import KSampling
from pymatgen.symmetry.bandstructure import HighSymmKpath
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from abipy.dfpt.anaddbnc import AnaddbNcFile
from abipy.abio.inputs import AnaddbInput
from abipy.flowtk.tasks import AnaddbTask, TaskManager
from abipy.dfpt.ddb import AnaddbError, DielectricTensorGenerator, DdbFile
from abipy.dfpt.phonons import PhononBands
from abipy.core.abinit_units import eV_to_THz
from maggma.builders import Builder
from maggma.core import Store

from emmet.core.phonon import (
    PhononWarnings,
    ThermodynamicProperties,
    AbinitPhonon,
    VibrationalEnergy,
)
from emmet.core.phonon import (
    PhononDos,
    PhononBandStructure,
    PhononWebsiteBS,
    Ddb,
    ThermalDisplacement,
)
from emmet.core.polar import DielectricDoc, BornEffectiveCharges, IRDielectric
from emmet.core.utils import jsanitize

SETTINGS = EmmetBuildSettings()


class PhononBuilder(Builder):
    def __init__(
        self,
        phonon_materials: Store,
        ddb_source: Store,
        phonon: Store,
        phonon_bs: Store,
        phonon_dos: Store,
        ddb_files: Store,
        th_disp: Store,
        phonon_website: Store,
        query: Optional[Dict] = None,
        manager: Optional[TaskManager] = None,
        symprec: float = SETTINGS.SYMPREC,
        angle_tolerance: float = SETTINGS.ANGLE_TOL,
        chunk_size=100,
        **kwargs,
    ):
        """
        Creates a set of collections for materials generating different kind of data
        from the phonon calculations.
        The builder requires the execution of the anaddb tool available in abinit.
        The parts that may contain large amount of data are split from the main
        document and store in separated collections. Notice that in these cases
        the size of a single document may be above the 16MB limit allowed by the
        standard MongoDB document.

        Args:
            phonon_materials (Store): source Store of phonon materials documents
                containing abinit_input and abinit_output.
            ddb_source (Store): source Store of ddb files. Matching the data in the materials Store.
            phonon (Store): target Store of the phonon properties
            phonon_bs (Store): target Store for the phonon band structure. The document may
                exceed the 16MB limit of a mongodb collection.
            phonon_dos (Store): target Store for the phonon DOS. The document may
                exceed the 16MB limit of a mongodb collection.
            ddb_files (Store): target Store of the DDB files. The document may
                exceed the 16MB limit of a mongodb collection.
            th_disp (Store): target Store of the data related to the generalized phonon DOS
                with the mean square displacement tensor. The document may exceed the 16MB
                limit of a mongodb collection.
            phonon_website (Store): target Store for the phonon band structure in the phononwebsite
                format. The document may exceed the 16MB limit of a mongodb collection.
            query (dict): dictionary to limit materials to be analyzed
            manager (TaskManager): an instance of the abipy TaskManager. If None it
                will be generated from user configuration.
            symprec (float): tolerance for symmetry finding when determining the
                band structure path.
            angle_tolerance (float): angle tolerance for symmetry finding when
                determining the band structure path.
        """

        self.phonon_materials = phonon_materials
        self.phonon = phonon
        self.ddb_source = ddb_source
        self.phonon_bs = phonon_bs
        self.phonon_dos = phonon_dos
        self.ddb_files = ddb_files
        self.th_disp = th_disp
        self.phonon_website = phonon_website
        self.query = query or {}
        self.symprec = symprec
        self.angle_tolerance = angle_tolerance
        self.chunk_size = chunk_size

        if manager is None:
            self.manager = TaskManager.from_user_config()
        else:
            self.manager = manager

        super().__init__(
            sources=[phonon_materials, ddb_source],
            targets=[phonon, phonon_bs, phonon_dos, ddb_files, th_disp, phonon_website],
            chunk_size=chunk_size,
            **kwargs,
        )

    def prechunk(self, number_splits: int):  # pragma: no cover
        """
        Gets all materials that need phonons

        Returns:
            generator of materials to extract phonon properties
        """

        # All relevant materials that have been updated since phonon props were last calculated
        q = dict(self.query)

        mats = self.phonon.newer_in(self.phonon_materials, exhaustive=True, criteria=q)

        N = ceil(len(mats) / number_splits)

        for mpid_chunk in grouper(mats, N):
            yield {"query": {self.phonon_materials.key: {"$in": list(mpid_chunk)}}}

    def get_items(self) -> Iterator[Dict]:
        """
        Gets all materials that need phonons

        Returns:
            generator of materials to extract phonon properties
        """

        self.logger.info("Phonon Builder Started")

        self.logger.info("Setting indexes")
        self.ensure_indexes()

        # All relevant materials that have been updated since phonon props were last calculated
        q = dict(self.query)

        mats = self.phonon.newer_in(self.phonon_materials, exhaustive=True, criteria=q)
        self.logger.info("Found {} new materials for phonon data".format(len(mats)))

        # list of properties queried from the results DB
        # basic information
        projection = {
            "mp_id": 1,
            "spacegroup.number": 1,
            "abinit_input": 1,  # input data
            "abinit_output.ddb_id": 1,  # file ids to be fetched
        }

        for m in mats:
            item = self.phonon_materials.query_one(
                properties=projection, criteria={self.phonon_materials.key: m}
            )

            # Read the DDB file and pass as an object. Do not write here since in case of parallel
            # execution each worker will write its own file.
            ddb_data = self.ddb_source.query_one(
                criteria={"_id": item["abinit_output"]["ddb_id"]}
            )
            if not ddb_data:
                self.logger.warning(
                    f"DDB file not found for file id {item['abinit_output']['ddb_id']}"
                )
                continue

            try:
                item["ddb_str"] = ddb_data["data"].decode("utf-8")
            except Exception:
                self.logger.warning(
                    f"could not extract DDB for file id {item['abinit_output']['ddb_id']}"
                )
                continue

            yield item

    def process_item(self, item: Dict) -> Optional[Dict]:
        """
        Generates the full phonon document from an item

        Args:
            item (dict): a dict extracted from the phonon calculations results.

        Returns:
            dict: a dict with the set of phonon data to be saved in the stores.
        """
        self.logger.debug("Processing phonon item for {}".format(item["mp_id"]))

        try:
            structure = Structure.from_dict(item["abinit_input"]["structure"])

            abinit_input_vars = self.abinit_input_vars(item)
            phonon_properties = self.get_phonon_properties(item)
            sr_break = self.get_sum_rule_breakings(item)
            ph_warnings = get_warnings(
                sr_break["asr"], sr_break["cnsr"], phonon_properties["ph_bs"]
            )
            if PhononWarnings.NEG_FREQ not in ph_warnings:
                thermodynamic, vibrational_energy = get_thermodynamic_properties(
                    phonon_properties["ph_dos"]
                )
            else:
                thermodynamic, vibrational_energy = None, None

            becs = None
            if phonon_properties["becs"] is not None:
                becs = BornEffectiveCharges(
                    material_id=item["mp_id"],
                    symmetrized_value=phonon_properties["becs"],
                    value=sr_break["becs_nosymm"],
                    cnsr_break=sr_break["cnsr"],
                )

            ap = AbinitPhonon.from_structure(
                structure=structure,
                meta_structure=structure,
                include_structure=True,
                material_id=item["mp_id"],
                cnsr_break=sr_break["cnsr"],
                asr_break=sr_break["asr"],
                warnings=ph_warnings,
                dielectric=phonon_properties["dielectric"],
                becs=becs,
                ir_spectra=phonon_properties["ir_spectra"],
                thermodynamic=thermodynamic,
                vibrational_energy=vibrational_energy,
                abinit_input_vars=abinit_input_vars,
            )

            phbs = PhononBandStructure(
                material_id=item["mp_id"],
                band_structure=phonon_properties["ph_bs"].as_dict(),
            )

            phws = PhononWebsiteBS(
                material_id=item["mp_id"],
                phononwebsite=phonon_properties["ph_bs"].as_phononwebsite(),
            )

            phdos = PhononDos(
                material_id=item["mp_id"],
                dos=phonon_properties["ph_dos"].as_dict(),
                dos_method=phonon_properties["ph_dos_method"],
            )

            ddb = Ddb(material_id=item["mp_id"], ddb=item["ddb_str"])

            th_disp = ThermalDisplacement(
                material_id=item["mp_id"],
                structure=structure,
                nsites=len(structure),
                nomega=phonon_properties["th_disp"]["nomega"],
                ntemp=phonon_properties["th_disp"]["ntemp"],
                temperatures=phonon_properties["th_disp"]["tmesh"].tolist(),
                frequencies=phonon_properties["th_disp"]["wmesh"].tolist(),
                gdos_aijw=phonon_properties["th_disp"]["gdos_aijw"].tolist(),
                amu=phonon_properties["th_disp"]["amu_symbol"],
                ucif_t=phonon_properties["th_disp"]["ucif_t"].tolist(),
                ucif_string_t300k=phonon_properties["th_disp"]["ucif_string_t300k"],
            )

            self.logger.debug("Item generated for {}".format(item["mp_id"]))

            d = dict(
                abiph=jsanitize(ap.model_dump(), allow_bson=True),
                phbs=jsanitize(phbs.model_dump(), allow_bson=True),
                phws=jsanitize(phws.model_dump(), allow_bson=True),
                phdos=jsanitize(phdos.model_dump(), allow_bson=True),
                ddb=jsanitize(ddb.model_dump(), allow_bson=True),
                th_disp=jsanitize(th_disp.model_dump(), allow_bson=True),
            )

            return d
        except Exception as error:
            self.logger.warning(
                "Error generating the phonon properties for {}: {}".format(
                    item["mp_id"], error
                )
            )
            return None

    def get_phonon_properties(self, item: Dict) -> Dict:
        """
        Extracts the phonon properties from the item
        """

        # the temp dir should still exist when using the objects as some readings are done lazily
        with tempfile.TemporaryDirectory() as workdir:
            structure = Structure.from_dict(item["abinit_input"]["structure"])

            self.logger.debug("Running anaddb in {}".format(workdir))

            ddb_path = os.path.join(workdir, "{}_DDB".format(item["mp_id"]))
            with open(ddb_path, "wt") as ddb_file:
                ddb_file.write(item["ddb_str"])

            ddb = DdbFile.from_string(item["ddb_str"])
            has_bec = ddb.has_bec_terms()
            has_epsinf = ddb.has_epsinf_terms()

            anaddb_input, labels_list = self.get_properties_anaddb_input(
                item,
                bs=True,
                dos="tetra",
                lo_to_splitting=has_bec,
                use_dieflag=has_epsinf,
            )
            task = self.run_anaddb(
                ddb_path=ddb_path, anaddb_input=anaddb_input, workdir=workdir
            )

            with task.open_phbst() as phbst_file, AnaddbNcFile(
                task.outpath_from_ext("anaddb.nc")
            ) as ananc_file:
                # phbst
                phbands = phbst_file.phbands
                if has_bec:
                    phbands.read_non_anal_from_file(phbst_file.filepath)
                symm_line_bands = self.get_pmg_bs(phbands, labels_list)  # type: ignore

                # ananc
                if has_bec and ananc_file.becs is not None:
                    becs = ananc_file.becs.values.tolist()
                else:
                    becs = None
                if has_epsinf and ananc_file.epsinf is not None:
                    e_electronic = ananc_file.epsinf.tolist()
                else:
                    e_electronic = None
                e_total = (
                    ananc_file.eps0.tolist() if ananc_file.eps0 is not None else None
                )
                if e_electronic and e_total:
                    e_ionic = (ananc_file.eps0 - ananc_file.epsinf).tolist()
                    dielectric = DielectricDoc.from_ionic_and_electronic(
                        ionic=e_ionic,
                        electronic=e_electronic,
                        material_id=item["mp_id"],
                        structure=structure,
                        deprecated=False,
                    )
                else:
                    dielectric = None

                # both
                if (
                    e_electronic
                    and e_total
                    and ananc_file.oscillator_strength is not None
                ):
                    die_gen = DielectricTensorGenerator.from_objects(
                        phbands, ananc_file
                    )

                    ir_tensor = IRDielectricTensor(
                        die_gen.oscillator_strength,
                        die_gen.phfreqs,
                        die_gen.epsinf,
                        die_gen.structure,
                    ).as_dict()
                    ir_spectra = IRDielectric(ir_dielectric_tensor=ir_tensor)
                else:
                    ir_spectra = None

            dos_method = "tetrahedron"
            with task.open_phdos() as phdos_file:
                complete_dos = phdos_file.to_pymatgen()
                msqd_dos = phdos_file.msqd_dos

                # if the integrated dos is not close enough to the expected value (3*N_sites) rerun the DOS using
                # gaussian integration
                integrated_dos = phdos_file.phdos.integral()[-1][1]
                nmodes = 3 * len(phdos_file.structure)

            if np.abs(integrated_dos - nmodes) / nmodes > 0.01:
                self.logger.warning(
                    "Integrated DOS {} instead of {} for {}. Recalculating with gaussian".format(
                        integrated_dos, nmodes, item["mp_id"]
                    )
                )
                with tempfile.TemporaryDirectory() as workdir_dos:
                    anaddb_input_dos, _ = self.get_properties_anaddb_input(
                        item,
                        bs=False,
                        dos="gauss",
                        lo_to_splitting=has_bec,
                        use_dieflag=has_epsinf,
                    )
                    task_dos = self.run_anaddb(
                        ddb_path=ddb_path,
                        anaddb_input=anaddb_input_dos,
                        workdir=workdir_dos,
                    )
                    with task_dos.open_phdos() as phdos_file:
                        complete_dos = phdos_file.to_pymatgen()
                        msqd_dos = phdos_file.msqd_dos
                dos_method = "gaussian"

            data = {
                "ph_dos": complete_dos,
                "ph_dos_method": dos_method,
                "ph_bs": symm_line_bands,
                "becs": becs,
                "ir_spectra": ir_spectra,
                "dielectric": dielectric,
                "th_disp": msqd_dos.get_json_doc(tstart=0, tstop=800, num=161),
            }

            return data

    def get_sum_rule_breakings(self, item: dict) -> dict:
        """
        Extracts the breaking of the acoustic and charge neutrality sum rules.
        Runs anaddb to get the values.
        """
        structure = Structure.from_dict(item["abinit_input"]["structure"])
        anaddb_input = AnaddbInput.modes_at_qpoint(
            structure, [0, 0, 0], asr=0, chneut=0
        )

        with tempfile.TemporaryDirectory() as workdir:
            ddb_path = os.path.join(workdir, "{}_DDB".format(item["mp_id"]))
            with open(ddb_path, "wt") as ddb_file:
                ddb_file.write(item["ddb_str"])

            ddb = DdbFile.from_string(item["ddb_str"])
            has_bec = ddb.has_bec_terms()

            task = self.run_anaddb(ddb_path, anaddb_input, workdir)

            if has_bec:
                with AnaddbNcFile(task.outpath_from_ext("anaddb.nc")) as ananc_file:
                    becs = ananc_file.becs
                    becs_val = becs.values.tolist() if becs else None
                    cnsr = np.max(np.abs(becs.sumrule)) if becs else None
            else:
                becs_val = None
                cnsr = None

            with task.open_phbst() as phbst_file:
                phbands = phbst_file.phbands

            # If the ASR breaking could not be identified. set it to None to signal the
            # missing information. This may trigger a warning.
            try:
                asr_breaking = phbands.asr_breaking(
                    units="cm-1", threshold=0.9, raise_on_no_indices=True
                )
                asr = asr_breaking.absmax_break
            except RuntimeError as e:
                self.logger.warning(
                    "Could not find the ASR breaking for {}. Error: {}".format(
                        item["mp_id"], e
                    )
                )
                asr = None

            breakings = {"cnsr": cnsr, "asr": asr, "becs_nosymm": becs_val}

        return breakings

    def run_anaddb(
        self, ddb_path: str, anaddb_input: AnaddbInput, workdir: str
    ) -> AnaddbTask:
        """
        Runs anaddb. Raise AnaddbError if the calculation couldn't complete

        Args:
            ddb_path (str): path to the DDB file
            anaddb_input (AnaddbInput): the input for anaddb
            workdir (str): the directory where the calculation is run
        Returns:
            An abipy AnaddbTask instance.
        """

        task = AnaddbTask.temp_shell_task(
            anaddb_input, ddb_node=ddb_path, workdir=workdir, manager=self.manager
        )

        # Run the task here.
        self.logger.debug("Start anaddb for {}".format(ddb_path))
        task.start_and_wait(autoparal=False)
        self.logger.debug("Finished anaddb for {}".format(ddb_path))

        report = task.get_event_report()
        if not report.run_completed:
            raise AnaddbError(task=task, report=report)

        self.logger.debug("anaddb succesful for {}".format(ddb_path))

        return task

    def get_properties_anaddb_input(
        self,
        item: dict,
        bs: bool = True,
        dos: str = "tetra",
        lo_to_splitting: bool = True,
        use_dieflag: bool = True,
    ) -> Tuple[AnaddbInput, Optional[List]]:
        """
        creates the AnaddbInput object to calculate the phonon properties.
        It also returns the list of qpoints labels for generating the PhononBandStructureSymmLine.

        Args:
            item: the item to process
            bs (bool): if True the phonon band structure will be calculated
            dos (str): if 'tetra' the DOS will be calculated with the tetrahedron method,
                if 'gauss' with gaussian smearing, if None the DOS will not be calculated
            lo_to_splitting (bool): contributions from the LO-TO splitting for the phonon
                BS will be calculated.
            use_dieflag (bool): the dielectric tensor will be calculated.
        """

        ngqpt = item["abinit_input"]["ngqpt"]
        q1shft = [(0, 0, 0)]

        structure = Structure.from_dict(item["abinit_input"]["structure"])

        # use all the corrections
        dipdip = 1
        asr = 2
        chneut = 1

        inp = AnaddbInput(structure, comment="ANADB input for phonon bands and DOS")

        inp.set_vars(
            ifcflag=1,
            ngqpt=np.array(ngqpt),
            q1shft=q1shft,
            nqshft=len(q1shft),
            asr=asr,
            chneut=chneut,
            dipdip=dipdip,
        )

        # Parameters for the dos.
        if dos == "tetra":
            # Use tetrahedra with dense dosdeltae (required to get accurate value of the integral)
            prtdos = 2
            dosdeltae = 9e-07  # Ha = 2 cm^-1
            ng2qppa = 200000
            ng2qpt = KSampling.automatic_density(structure, kppa=ng2qppa).kpts[0]
            inp.set_vars(prtdos=prtdos, dosdeltae=dosdeltae, ng2qpt=ng2qpt)
        elif dos == "gauss":
            # Use gauss with denser grid and a smearing
            prtdos = 1
            dosdeltae = 4.5e-06  # Ha = 10 cm^-1
            ng2qppa = 500000
            dossmear = 1.82e-5  # Ha = 4 cm^-1
            ng2qpt = KSampling.automatic_density(structure, kppa=ng2qppa).kpts[0]
            inp.set_vars(
                prtdos=prtdos, dosdeltae=dosdeltae, dossmear=dossmear, ng2qpt=ng2qpt
            )
        elif dos is not None:
            raise ValueError("Unsupported value of dos.")

        # Parameters for the BS
        labels_list = None
        if bs:
            spga = SpacegroupAnalyzer(
                structure, symprec=self.symprec, angle_tolerance=self.angle_tolerance
            )

            spgn = spga.get_space_group_number()
            if spgn != item["spacegroup"]["number"]:
                raise RuntimeError(
                    "Parsed specegroup number {} does not match "
                    "calculation spacegroup {}".format(
                        spgn, item["spacegroup"]["number"]
                    )
                )

            hs = HighSymmKpath(
                structure, symprec=self.symprec, angle_tolerance=self.angle_tolerance
            )

            qpts, labels_list = hs.get_kpoints(
                line_density=18, coords_are_cartesian=False
            )

            n_qpoints = len(qpts)
            qph1l = np.zeros((n_qpoints, 4))

            qph1l[:, :-1] = qpts
            qph1l[:, -1] = 1

            inp["qph1l"] = qph1l.tolist()
            inp["nph1l"] = n_qpoints

            if lo_to_splitting:
                kpath = hs.kpath
                directions = []  # type: list
                for qptbounds in kpath["path"]:
                    for i, qpt in enumerate(qptbounds):
                        if np.array_equal(kpath["kpoints"][qpt], (0, 0, 0)):
                            # anaddb expects cartesian coordinates for the qph2l list
                            if i > 0:
                                directions.extend(
                                    structure.lattice.reciprocal_lattice_crystallographic.get_cartesian_coords(
                                        kpath["kpoints"][qptbounds[i - 1]]
                                    )
                                )
                                directions.append(0)

                            if i < len(qptbounds) - 1:
                                directions.extend(
                                    structure.lattice.reciprocal_lattice_crystallographic.get_cartesian_coords(
                                        kpath["kpoints"][qptbounds[i + 1]]
                                    )
                                )
                                directions.append(0)

                if directions:
                    directions = np.reshape(directions, (-1, 4))  # type: ignore
                    inp.set_vars(nph2l=len(directions), qph2l=directions)

        # Parameters for dielectric constant
        if use_dieflag:
            inp["dieflag"] = 1

        return inp, labels_list

    @staticmethod
    def get_pmg_bs(
        phbands: PhononBands, labels_list: List
    ) -> PhononBandStructureSymmLine:
        """
        Generates a PhononBandStructureSymmLine starting from a abipy PhononBands object

        Args:
            phbands (PhononBands): the phonon band structures
            labels_list (list): list of labels used to generate the path
        Returns:
            An instance of PhononBandStructureSymmLine
        """

        structure = phbands.structure

        n_at = len(structure)

        qpts = np.array(phbands.qpoints.frac_coords)
        ph_freqs = np.array(phbands.phfreqs)
        displ = np.array(phbands.phdispl_cart)

        labels_dict = {}

        for i, (q, l) in enumerate(zip(qpts, labels_list)):
            if l:
                labels_dict[l] = q
                # set LO-TO at gamma
                if phbands.non_anal_ph and "Gamma" in l:
                    if i > 0 and not labels_list[i - 1]:
                        ph_freqs[i] = phbands._get_non_anal_freqs(qpts[i - 1])
                        displ[i] = phbands._get_non_anal_phdispl(qpts[i - 1])
                    if i < len(qpts) - 1 and not labels_list[i + 1]:
                        ph_freqs[i] = phbands._get_non_anal_freqs(qpts[i + 1])
                        displ[i] = phbands._get_non_anal_phdispl(qpts[i + 1])

        ph_freqs = np.transpose(ph_freqs) * eV_to_THz
        displ = np.transpose(
            np.reshape(displ, (len(qpts), 3 * n_at, n_at, 3)), (1, 0, 2, 3)
        )

        ph_bs_sl = PhononBandStructureSymmLine(
            qpoints=qpts,
            frequencies=ph_freqs,
            lattice=structure.reciprocal_lattice,
            has_nac=phbands.non_anal_ph is not None,
            eigendisplacements=displ,
            labels_dict=labels_dict,
            structure=structure,
        )

        ph_bs_sl.band_reorder()

        return ph_bs_sl

    @staticmethod
    def abinit_input_vars(item: dict) -> dict:
        """
        Extracts the useful abinit input parameters from an item.
        """

        i = item["abinit_input"]

        data = {}

        def get_vars(label):
            if label in i and i[label]:
                return {k: v for (k, v) in i[label]["abi_args"]}
            else:
                return {}

        data["gs_input"] = get_vars("gs_input")
        data["ddk_input"] = get_vars("ddk_input")
        data["dde_input"] = get_vars("dde_input")
        data["phonon_input"] = get_vars("phonon_input")
        data["wfq_input"] = get_vars("wfq_input")

        data["ngqpt"] = i["ngqpt"]
        data["ngkpt"] = i["ngkpt"]
        data["shiftk"] = i["shiftk"]
        data["ecut"] = i["ecut"]
        data["occopt"] = i["occopt"]
        data["tsmear"] = i.get("tsmear", 0)

        data["pseudopotentials"] = {
            "name": i["pseudopotentials"]["pseudos_name"],
            "md5": i["pseudopotentials"]["pseudos_md5"],
        }

        return data

    def update_targets(self, items: List[Dict]):
        """
        Inserts the new task_types into the task_types collection

        Args:
            items ([dict]): a list of phonon dictionaries to update
        """
        items = list(filter(None, items))
        items_ph = [i["abiph"] for i in items]
        items_ph_band = [i["phbs"] for i in items]
        items_ph_dos = [i["phdos"] for i in items]
        items_ddb = [i["ddb"] for i in items]
        items_th_disp = [i["th_disp"] for i in items]
        items_ph_web = [i["phws"] for i in items]

        if len(items) > 0:
            self.logger.info("Updating {} phonon docs".format(len(items)))
            self.phonon.update(docs=items_ph)
            self.phonon_bs.update(docs=items_ph_band)
            self.phonon_dos.update(docs=items_ph_dos)
            self.ddb_files.update(docs=items_ddb)
            self.th_disp.update(docs=items_th_disp)
            self.phonon_website.update(docs=items_ph_web)

        else:
            self.logger.info("No items to update")

    def ensure_indexes(self):
        """
        Ensures indexes on the tasks and materials collections
        """
        self.phonon_materials.ensure_index(self.phonon_materials.key, unique=True)

        self.phonon.ensure_index(self.phonon.key, unique=True)
        self.phonon_bs.ensure_index(self.phonon.key, unique=True)
        self.phonon_dos.ensure_index(self.phonon.key, unique=True)
        self.ddb_files.ensure_index(self.phonon.key, unique=True)
        self.th_disp.ensure_index(self.phonon.key, unique=True)
        self.phonon_website.ensure_index(self.phonon.key, unique=True)


def get_warnings(
    asr_break: float, cnsr_break: float, ph_bs: PhononBandStructureSymmLine
) -> List[PhononWarnings]:
    """

    Args:
        asr_break (float): the largest breaking of the acoustic sum rule in cm^-1
        cnsr_break (float): the largest breaking of the charge neutrality sum rule
        ph_bs (PhononBandStructureSymmLine): the phonon band structure

    Returns:
        PhononWarnings: the model containing the data of the warnings.
    """

    warnings = []

    if asr_break and asr_break > 30:
        warnings.append(PhononWarnings.ASR)
    if cnsr_break and cnsr_break > 0.2:
        warnings.append(PhononWarnings.CNSR)

    # neglect small negative frequencies (0.03 THz ~ 1 cm^-1)
    limit = -0.03

    bands = np.array(ph_bs.bands)
    neg_freq = bands < limit

    # there are negative frequencies anywhere in the BZ
    if np.any(neg_freq):
        warnings.append(PhononWarnings.NEG_FREQ)

        qpoints = np.array([q.frac_coords for q in ph_bs.qpoints])

        qpt_has_neg_freq = np.any(neg_freq, axis=0)

        if np.max(np.linalg.norm(qpoints[qpt_has_neg_freq], axis=1)) < 0.05:
            warnings.append(PhononWarnings.SMALL_Q_NEG_FREQ)

    return warnings


def get_thermodynamic_properties(
    ph_dos: CompletePhononDos,
) -> Tuple[ThermodynamicProperties, VibrationalEnergy]:
    """
    Calculates the thermodynamic properties from a phonon DOS

    Args:
        ph_dos (CompletePhononDos): The DOS used to calculate the properties.

    Returns:
        ThermodynamicProperties and VibrationalEnergy: the models containing the calculated thermodynamic
            properties and vibrational contribution to the total energy.
    """

    tstart, tstop, nt = 0, 800, 161
    temp = np.linspace(tstart, tstop, nt)

    cv = []
    entropy = []
    internal_energy = []
    helmholtz_free_energy = []

    for t in temp:
        cv.append(ph_dos.cv(t, ph_dos.structure))
        entropy.append(ph_dos.entropy(t, ph_dos.structure))
        internal_energy.append(ph_dos.internal_energy(t, ph_dos.structure))
        helmholtz_free_energy.append(ph_dos.helmholtz_free_energy(t, ph_dos.structure))

    zpe = ph_dos.zero_point_energy(ph_dos.structure)

    temperatures = temp.tolist()
    tp = ThermodynamicProperties(temperatures=temperatures, cv=cv, entropy=entropy)

    ve = VibrationalEnergy(
        temperatures=temperatures,
        internal_energy=internal_energy,
        helmholtz_free_energy=helmholtz_free_energy,
        zero_point_energy=zpe,
    )

    return tp, ve



from pathlib import Path
from emmet.builders.settings import EmmetBuildSettings

import pytest
from maggma.stores import JSONStore, MemoryStore
from monty.serialization import dumpfn, loadfn

from emmet.builders.vasp.materials import MaterialsBuilder
from emmet.builders.vasp.task_validator import TaskValidator


@pytest.fixture(scope="session")
def tasks_store(test_dir):
    return JSONStore(test_dir / "test_si_tasks.json.gz")


@pytest.fixture(scope="session")
def validation_store(tasks_store):
    settings = EmmetBuildSettings(VASP_VALIDATE_POTCAR_STATS=False)
    validation_store = MemoryStore()
    builder = TaskValidator(
        tasks=tasks_store, task_validation=validation_store, settings=settings
    )
    builder.run()
    return validation_store


@pytest.fixture
def materials_store():
    return MemoryStore()


def test_materials_builder(tasks_store, validation_store, materials_store):
    builder = MaterialsBuilder(
        tasks=tasks_store, task_validation=validation_store, materials=materials_store
    )
    builder.run()
    assert materials_store.count() == 1
    assert materials_store.count({"deprecated": False}) == 1


def test_serialization(tmpdir):
    builder = MaterialsBuilder(MemoryStore(), MemoryStore(), MemoryStore())

    dumpfn(builder.as_dict(), Path(tmpdir) / "test.json")
    loadfn(Path(tmpdir) / "test.json")
